/* $Id: pm2phil.c,v 1.2 2000/06/20 23:48:58 becka Exp $
******************************************************************************

   Vidio driver for the 3dlabs Permedia 2 using Phillips SAA 7111A+7124.
   This chipset combination is used e.g. in the ELSA Winner Office 2000.

   Copyright (C) 2000 Andreas Beck	[becka@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#define	DRIVER_NAME	"Permedia 2 videoio driver"
#define	DRIVER_REV	"$Revision: 1.2 $"

/*
**	KGI includes
*/

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>

#include <3dlabs/pm2free.h>

/* Standard PM2 function.
 */
static void wait_pm2(struct chipset_private* i) {

	WAIT_FIFO(i, 1);
	pm2_WR(i, PM2R_SYNC, 0);
	DEFRW();
	do {
		while (pm2_RD(i, PM2R_OUT_FIFO_WORDS)==0);
		DEFR();
	} while (pm2_RD(i, PM2R_OUT_FIFO)!=PM2TAG(PM2R_SYNC));
}

/* First include the lowlevel I2C routines for the Permedia2 chipset.
 */

#include "pm2_I2C_ll.inc"

/* Now the generic Bitlevel->I2C Highlevel stuff.
 */

#include "I2C_hl.inc"

/* Now the function interfaces to the two chips.
 */

#include "SAA7111A.inc"
#include "SAA7124.inc"

/*****************************************************************************/
/* PM2 part of driver starts here. */
/*****************************************************************************/

static void PM2VidInDisable(struct chipset_private* priv) {
	pm2_WR(priv,0x5900,(pm2_RD(priv,0x5900)&~0x03));
}

static void PM2VidInEnable(struct chipset_private* priv,
				kgi_vidio_reqmode *reqmode, kgiu32 *offsets) {

	int xfactor,yfactor;
	int width=reqmode->visible.x;
	int virtwidth=reqmode->virt.x;
	int height=reqmode->visible.y;

	pm2_WR(priv,0x5968,0);
	pm2_WR(priv,0x5970,0);
	pm2_WR(priv,0x5978,0);
	pm2_WR(priv,0x5980,0);
	pm2_WR(priv,0x5988,0);
	pm2_WR(priv,0x5990,0);
	pm2_WR(priv,0x5998,0);
	pm2_WR(priv,0x59a0,0);
	pm2_WR(priv,0x59a8,0);
	pm2_WR(priv,0x59b0,0);
	pm2_WR(priv,0x5920,0);
	pm2_WR(priv,0x5918,0x0003);
	pm2_WR(priv,0x5900,(pm2_RD(priv,0x5900)&~0x00000004)|0x00000004);
	pm2_WR(priv,0x5908,0x7ff);
	pm2_WR(priv,0x5900,(pm2_RD(priv,0x5900)&~0x00000100)|0x00000000);
	pm2_WR(priv,0x5940,virtwidth/4);
	pm2_WR(priv,0x5928,offsets[0]/8);
	pm2_WR(priv,0x5930,offsets[1]/8);
	pm2_WR(priv,0x5938,offsets[2]/8);
// x-mirror	pm2_WR(priv,0x5900,(pm2_RD(priv,0x5900)&~0x00000080)|0x00000080);
	pm2_WR(priv,0x5900,(pm2_RD(priv,0x5900)&~0x00000080)|0x00000000);

	xfactor=yfactor=0;
	while(height<=144) {
		height*=2;
		yfactor++;
	}
	while(height>288) {	/* Interlace */
		height/=2;
		yfactor--;
	}	/* Height is now in [145..288]*/
	while(width<=352) {
		width*=2;
		xfactor++;
	}
	xfactor<<=3;
	if (yfactor>=0)
		yfactor<<=5;
	else
		yfactor=0x800;

	/* Here ELSA calls one of PM2VidInSetupSub1-9 - see oldcode.h
	 * We use a more generic version that calculates the parameters
	 */
	pm2_WR(priv,0x59b8,0x84);
	pm2_WR(priv,0x5800, (pm2_RD(priv,0x5800)&~0x0003fe00)|0x00020c00 );
	pm2_WR(priv,0x5900, (pm2_RD(priv,0x5900)&~0x00001e78)|xfactor|yfactor );
	pm2_WR(priv,0x5948,reqmode->offset.y+0xa0-height/2);
	pm2_WR(priv,0x5950,reqmode->offset.y+0xa0+height/2);
	pm2_WR(priv,0x5958,reqmode->offset.x+0x03f0-width);
	pm2_WR(priv,0x5960,reqmode->offset.x+0x03f0+width);
	pm2_WR(priv,0x5900, (pm2_RD(priv,0x5900)&~0x00000001)|0x00000001 );

//	pm2_WR(priv,0x3078,0x1107);	// Fixup timing to use on Burstsize=8. 
}

static void PM2VidOutEnable(struct chipset_private* priv,
				kgi_vidio_reqmode *reqmode, kgiu32 *offsets) {

	int width	=reqmode->visible.x;
	int height	=reqmode->visible.y;
	int virtwidth	=reqmode->virt.x;
	int is16	=(reqmode->graphtype==KGIGT_16BIT);
	int bytewidth	=2;
	

	switch(reqmode->graphtype) {
		case KGIGT_16BIT:
			bytewidth=2;break;
		case KGIGT_32BIT:
			bytewidth=4;
			break;
		default:
			TRACE(printk("What ? Illegal graphtype selected. \n"));
			break;
	}


	if (!is16&&(KGIGT_32BIT!=reqmode->graphtype)) return;

	pm2_WR(priv,0x5a08,0x07ff);
	if (is16)
		pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x00000008)|0x00000000);
	else
		pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x00000008)|0x00000008);
	pm2_WR(priv,0x5ab8,0x0400);
	pm2_WR(priv,0x5a70,0x0000);
	pm2_WR(priv,0x5a68,0x0003);
	pm2_WR(priv,0x5a78,0x0000);
	pm2_WR(priv,0x5a80,0x0000);
	pm2_WR(priv,0x5a88,0x0000);
	pm2_WR(priv,0x5a90,0x0000);
	pm2_WR(priv,0x5a98,0x0008);
	pm2_WR(priv,0x5aa0,0x0008);
	pm2_WR(priv,0x5aa8,0x0008);
	pm2_WR(priv,0x5ab0,0x0008);
	pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x00000004)|0x00000004);
	pm2_WR(priv,0x5a20,0x0000);
	pm2_WR(priv,0x5a18,0x0003);

	pm2_WR(priv,0x5a40,virtwidth/8 /* virtual width */);
	pm2_WR(priv,0x5a28,offsets[0]/8);	/* 3x picture position */
	pm2_WR(priv,0x5a30,offsets[1]/8);
	pm2_WR(priv,0x5a38,offsets[2]/8);

	pm2_WR(priv,0x5800,(pm2_RD(priv,0x5800)&~0x0bfc0000)|0x00180000);

	pm2_WR(priv,0x5a48,reqmode->offset.y+150-height/4);	/* y start - top margin */
	pm2_WR(priv,0x5a50,reqmode->offset.y+150+height/4);	/* y stop - height */
	pm2_WR(priv,0x5a58,reqmode->offset.x*bytewidth+1012-width);		/* x start - left margin */
	pm2_WR(priv,0x5a60,reqmode->offset.x*bytewidth+1012+width);		/* x stop - width */

	if (is16)
		pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x000007f0)|0x00000210);
	else
		pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x000007f0)|0x00000400);
	pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x00000801)|0x00000801);

//	pm2_WR(priv,0x3078,0x11); 	/* enable loose RAM timing */
}

static void PM2VidOutDisable(struct chipset_private* priv) {
	pm2_WR(priv,0x5a00,(pm2_RD(priv,0x5a00)&~0x03));
}

static void PM2VidIOInit(struct chipset_private* priv) {
	PM2VidInDisable(priv);
	PM2VidOutDisable(priv);
	pm2_WR(priv,0x5800,(pm2_RD(priv,0x5800)&~0x07)|0x03);
}

static void PM2VidIODisable(struct chipset_private* priv) {
	PM2VidInDisable(priv);
	PM2VidOutDisable(priv);
	pm2_WR(priv,0x5800,0x01f3);
}

/* These are due to the combination of both SAA chips. When both are active,
 * they need to be locked together, or trouble will occur.
 */

static void SAA_locktogether(struct chipset_private* priv) {

	SAA7111A_setreg(priv, 0x08,(SAA7111A_getreg(priv, 0x08)&~4)|8 );	/* Close PLL - VTR mode */
#if 0
	SAA7124_setreg(priv,0x5e,SAA7124_getshadowreg(priv, 0x5e)|0x80 );	/* RTCI from SAA7111 */
	SAA7124_setreg(priv,0x62,SAA7124_getshadowreg(priv, 0x62)|0x80 );	/* RTCE from SAA7111 */
#else
	SAA7124_setreg(priv,0x5e,SAA7124_getshadowreg(priv, 0x5e)&~0x80 );	/* no RTCI from SAA7111 */
	SAA7124_setreg(priv,0x62,SAA7124_getshadowreg(priv, 0x62)&~0x80 );	/* no RTCE from SAA7111 */
#endif
}

static void SAA_unlock(struct chipset_private* priv) {
	SAA7111A_setreg(priv,0x08,SAA7111A_getreg(priv, 0x08)|0x0c );	/* Close PLL - VTR mode */
	SAA7124_setreg (priv,0x5e,SAA7124_getshadowreg(priv, 0x5e)&~0x80 );	/* no RTCI from SAA7111 */
	SAA7124_setreg (priv,0x62,SAA7124_getshadowreg(priv, 0x62)&~0x80 );	/* no RTCE from SAA7111 */
}

static void VidIOInitializeAll(struct chipset_private* priv) {
	I2CStop(priv);
	SAA7111A_getreg(priv,0x00);	/* Clear the bus */
	SAA7111A_setup1(priv);		/* CVBS PAL */
	SAA7111A_imgctrl(priv,0x80,0x47,0x40,0x00,0x01);
	SAA7124_setup(priv);
	SAA7124_testpattern(priv);	/* Testpattern Generator. */
	SAA7111A_basic(priv);
}

static void VidIODeinitializeAll(struct chipset_private* priv) {	// 0001.76bc
	SAA7124_disable(priv);		/* shut down 7124 */
	SAA7124_testpattern(priv);	/* Testpattern Generator */
	SAA7111A_basic(priv);	
	SAA_unlock(priv);
}

/*****************************************************************************/


kgi_vidio_reqmode vmode_in = {
	3,
	{ 352,288 },
	{ 352,288 },
	{ 0,0 },
	{ 2,2 },
	KGIGT_16BIT
};
kgiu32 voffsets_in[3]={0,0,0};

kgi_vidio_reqmode vmode_out= {
	3,
	{ 640,576 },
	{ 640,576 },
	{ 0,0 },
	{ 2,2 },
	KGIGT_16BIT
};
kgiu32 voffsets_out[3]={0,0,0};

static kgi_vidio_avail enable_in=0,enable_out=0;


/*
**	KGI module interface
*/

int kgim_vidio_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct chipset_private *priv = KGIM_PRIVATE(dpy, chipset);

	/* FIXME ! Probe for chips ! */
	PM2VidIODisable(priv);
	PM2VidIOInit(priv);
	VidIOInitializeAll(priv);
	SAA7124_setup(priv);
	SAA_locktogether(priv);
	SAA7124_testpattern(priv);	// Testpattern Generator. 
	SAA7124_setup(priv);
	SAA7124_testpattern(priv);	// Testpattern Generator. 
	PM2VidInDisable(priv);
	SAA_unlock(priv);

//	PM2VidOutEnable(priv,&vmode_out,voffsets_out);

	TRACE(printk("pm2phil vidio driver loaded\n"));

	return EOK;
}

void kgim_vidio_done(struct kgi_display *dpy)
{
	struct chipset_private *priv = KGIM_PRIVATE(dpy, chipset);

	VidIODeinitializeAll(priv);

	TRACE(printk("pm2phil vidio driver removed\n"));
}

int kgim_vidio_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			  enum kgi_tm_cmd *cmd)
{
	return EOK;
}

void kgim_vidio_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	/* !!! We should set up the accel here !!! */
	return;
}
int vidout_checkmode(kgi_vidio_reqmode *mode) {
	int res=0;
	int bytewidth=1;

	if (mode->frames!=3) {
		res=(mode->frames!=0);
		mode->frames=3;
	}
	if (mode->scale.x!=1) {
		res=(mode->scale.x!=0);
		mode->scale.x=1;
	}
	if (mode->scale.y!=1) {
		res=(mode->scale.y!=0);
		mode->scale.y=1;
	}

	switch(mode->graphtype) {
		case KGIGT_16BIT:
			bytewidth=2;
			break;
		case KGIGT_32BIT:
			bytewidth=4;
			break;
		default:
			res=(mode->graphtype!=0);
			mode->graphtype=KGIGT_16BIT;
			break;
	}

	if (mode->visible.x==0) 
		mode->visible.x=mode->visible.y*4/3;
	if (mode->visible.y==0) 
		mode->visible.y=mode->visible.x*3/4;
	if (mode->visible.x==0) {
		mode->visible.x=352;
		mode->visible.y=288;
	}
	if (mode->visible.x>704) {
		mode->visible.x=704;res=1;
	}
	if (mode->visible.y>576) {
		mode->visible.y=576;res=1;
	}

	if (mode->virt.x<mode->visible.x*bytewidth) {
		res=(mode->virt.x!=0);
		mode->virt.x=mode->visible.x*bytewidth;
	}
	if (mode->virt.y<mode->visible.y*bytewidth) {
		res=(mode->virt.y!=0);
		mode->virt.y=mode->visible.y*bytewidth;
	}
	return res;	
}

int vidin_checkmode(kgi_vidio_reqmode *mode) {
	int res=0;

	if (mode->frames!=3) {
		res=(mode->frames!=0);
		mode->frames=3;
	}
	if (mode->scale.x!=1) {
		res=(mode->scale.x!=0);
		mode->scale.x=1;
	}
	if (mode->scale.y!=1) {
		res=(mode->scale.y!=0);
		mode->scale.y=1;
	}

	switch(mode->graphtype) {
		case KGIGT_16BIT:
			break;
		default:
			res=(mode->graphtype!=0);
			mode->graphtype=KGIGT_16BIT;
			break;
	}

	switch(mode->colmodel) {
		case KGICM_YUV:
			break;
		default:
			res=(mode->colmodel!=0);
			mode->graphtype=KGICM_YUV;
			break;
	}

	if (mode->visible.x==0) 
		mode->visible.x=mode->visible.y*4/3;
	if (mode->visible.y==0) 
		mode->visible.y=mode->visible.x*3/4;
	if (mode->visible.x==0) {
		mode->visible.x=352;
		mode->visible.y=288;
	}
	if (mode->visible.x>704) {
		mode->visible.x=704;res=1;
	}
	if (mode->visible.y>576) {
		mode->visible.y=576;res=1;
	}

	if (mode->virt.x<mode->visible.x) {
		res=(mode->virt.x!=0);
		mode->virt.x=mode->visible.x;
	}
	if (mode->virt.y<mode->visible.y) {
		res=(mode->virt.y!=0);
		mode->virt.y=mode->visible.y;
	}
	return res;	
}

int kgim_vidio_command(struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{ 
	struct chipset_private *priv = KGIM_PRIVATE(gr->dev.dpy, chipset);

	switch (cmd) {
		case VIDIO_IN_TESTAVAIL:
		case VIDIO_OUT_TESTAVAIL:
			gr->io_res = gr->io_buf;
			*((kgi_vidio_avail*) gr->io_res)=KGI_VIDIO_AVAIL_ENABLE;
			return 0;
		case VIDIO_OUT_ENABLE: {
				kgi_vidio_avail new;
				new=*((kgi_vidio_avail*) gr->io_buf);
				if (new==enable_out) return 0;
				if ( (new       &KGI_VIDIO_AVAIL_ENABLE) !=
				     (enable_out&KGI_VIDIO_AVAIL_ENABLE) ) {
					if (new&KGI_VIDIO_AVAIL_ENABLE) {
						SAA7124_setup(priv);
						PM2VidOutEnable(priv,&vmode_out,voffsets_out);
					} else {
						PM2VidOutDisable(priv);
						SAA7124_testpattern(priv);
					}
				}
				enable_out=new;
			}
		case VIDIO_OUT_CHECKMODE:
			gr->io_res = gr->io_buf;
			vidout_checkmode(gr->io_res);
			return 0;
		case VIDIO_OUT_SETMODE:
			gr->io_res = gr->io_buf;
			vidout_checkmode(gr->io_res);
			memcpy(&vmode_out,gr->io_res,sizeof(vmode_out));
			if (enable_out&KGI_VIDIO_AVAIL_ENABLE) {
				PM2VidOutEnable(priv,&vmode_out,voffsets_out);
			}
			return 0;
		case VIDIO_OUT_GETMODE:
			gr->io_res = gr->io_buf;
			memcpy(gr->io_res,&vmode_out,sizeof(vmode_out));
			return 0;
		case VIDIO_OUT_SETFRAMESTART:{
			kgi_vidio_framestart *fs;
			fs=gr->io_buf;
			if (fs->frame<3) {
				voffsets_out[fs->frame]=fs->offset;
				return 0;
			}
		}	break;
		case VIDIO_OUT_GETFRAMESTART:{
			kgi_vidio_framestart *fs;
			fs=gr->io_res=gr->io_buf;
			if (fs->frame<3) {
				fs->offset=voffsets_out[fs->frame];
				return 0;
			}
		}	break;
		case VIDIO_IN_ENABLE: {
				kgi_vidio_avail new;
				new=*((kgi_vidio_avail*) gr->io_buf);
				if (new==enable_in) return 0;
				if ( (new       &KGI_VIDIO_AVAIL_ENABLE) !=
				     (enable_in &KGI_VIDIO_AVAIL_ENABLE) ) {
					if (new&KGI_VIDIO_AVAIL_ENABLE) {
						SAA_locktogether(priv);
						PM2VidInEnable(priv,&vmode_in,voffsets_in);
					} else {
						PM2VidInDisable(priv);
						SAA_unlock(priv);
					}
				}
				enable_in=new;
			}
		case VIDIO_IN_CHECKMODE:
			gr->io_res = gr->io_buf;
			vidin_checkmode(gr->io_res);
			return 0;
		case VIDIO_IN_SETMODE:
			gr->io_res = gr->io_buf;
			vidin_checkmode(gr->io_res);
			memcpy(&vmode_in,gr->io_res,sizeof(vmode_in));
			if (enable_in&KGI_VIDIO_AVAIL_ENABLE) {
				PM2VidInEnable(priv,&vmode_in,voffsets_in);
			}
			return 0;
		case VIDIO_IN_GETMODE:
			gr->io_res = gr->io_buf;
			memcpy(gr->io_res,&vmode_in,sizeof(vmode_in));
			return 0;
		case VIDIO_IN_SETFRAMESTART:{
			kgi_vidio_framestart *fs;
			fs=gr->io_buf;
			if (fs->frame<3) {
				voffsets_in[fs->frame]=fs->offset;
				return 0;
			}
		}	break;
		case VIDIO_IN_GETFRAMESTART:{
			kgi_vidio_framestart *fs;
			fs=gr->io_res=gr->io_buf;
			if (fs->frame<3) {
				fs->offset=voffsets_in[fs->frame];
				return 0;
			}
		}	break;
		default:
			break;
	}
			
	return -E(VIDIO, UNKNOWN);
}
